import asyncio
import socket
import time
from contextlib import closing
from functools import wraps
from importlib import import_module
from pathlib import Path
from threading import Event, Thread
from typing import Any, Callable, List, Optional

from typing_extensions import ParamSpec

import idom

from .proto import ServerFactory


CLIENT_BUILD_DIR = Path(idom.__file__).parent / "client"

_SUPPORTED_PACKAGES = [
    "sanic",
    "fastapi",
    "flask",
    "tornado",
    "starlette",
]


_FuncParams = ParamSpec("_FuncParams")


def threaded(function: Callable[_FuncParams, None]) -> Callable[_FuncParams, Thread]:
    @wraps(function)
    def wrapper(*args: Any, **kwargs: Any) -> Thread:
        def target() -> None:
            asyncio.set_event_loop(asyncio.new_event_loop())
            function(*args, **kwargs)

        thread = Thread(target=target, daemon=True)
        thread.start()

        return thread

    return wrapper


def wait_on_event(description: str, event: Event, timeout: Optional[float]) -> None:
    if not event.wait(timeout):
        raise TimeoutError(f"Did not {description} within {timeout} seconds")


def poll(
    description: str,
    frequency: float,
    timeout: Optional[float],
    function: Callable[[], bool],
) -> None:
    if timeout is not None:
        expiry = time.time() + timeout
        while not function():
            if time.time() > expiry:
                raise TimeoutError(f"Did not {description} within {timeout} seconds")
            time.sleep(frequency)
    else:
        while not function():
            time.sleep(frequency)


def find_builtin_server_type(type_name: str) -> ServerFactory[Any, Any]:
    """Find first installed server implementation

    Raises:
        :class:`RuntimeError` if one cannot be found
    """
    installed_builtins: List[str] = []
    for name in _SUPPORTED_PACKAGES:
        try:
            import_module(name)
        except ImportError:  # pragma: no cover
            continue
        else:
            builtin_module = import_module(f"idom.server.{name}")
            installed_builtins.append(builtin_module.__name__)
        try:
            return getattr(builtin_module, type_name)  # type: ignore
        except AttributeError:  # pragma: no cover
            pass
    else:  # pragma: no cover
        if not installed_builtins:
            raise RuntimeError(
                f"Found none of the following builtin server implementations {_SUPPORTED_PACKAGES}"
            )
        else:
            raise ImportError(
                f"No server type {type_name!r} found in installed implementations {installed_builtins}"
            )


def find_available_port(
    host: str,
    port_min: int = 8000,
    port_max: int = 9000,
    allow_reuse_waiting_ports: bool = True,
) -> int:
    """Get a port that's available for the given host and port range"""
    for port in range(port_min, port_max):
        with closing(socket.socket()) as sock:
            try:
                if allow_reuse_waiting_ports:
                    # As per this answer: https://stackoverflow.com/a/19247688/3159288
                    # setting can be somewhat unreliable because we allow the use of
                    # ports that are stuck in TIME_WAIT. However, not setting the option
                    # means we're overly cautious and almost always use a different addr
                    # even if it could have actually been used.
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.bind((host, port))
            except OSError:
                pass
            else:
                return port
    raise RuntimeError(
        f"Host {host!r} has no available port in range {port_max}-{port_max}"
    )



from typing import Any, Dict, Sequence
from warnings import warn


def _wrap_in_warning_event_proxies(values: Sequence[Any]) -> Sequence[Any]:
    return [_EventProxy(x) if isinstance(x, dict) else x for x in values]


class _EventProxy(Dict[Any, Any]):
    def __getitem__(self, key: Any) -> Any:  # pragma: no cover
        try:
            return super().__getitem__(key)
        except KeyError:
            target = self.get("target")
            if isinstance(target, dict) and key in target:
                warn(
                    f"The event key event[{key!r}] has been moved event['target'][{key!r}",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return target[key]
            else:
                raise

    def get(self, key: Any, default: Any = None) -> Any:  # pragma: no cover
        try:
            return super().__getitem__(key)
        except KeyError:
            target = self.get("target")
            if isinstance(target, dict) and key in target:
                warn(
                    f"The event key event[{key!r}] has been moved event['target'][{key!r}",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return target[key]
            else:
                return default



import logging
import re
from pathlib import Path, PurePosixPath
from typing import Set, Tuple
from urllib.parse import urlparse

import requests


logger = logging.getLogger(__name__)


def module_name_suffix(name: str) -> str:
    if name.startswith("@"):
        name = name[1:]
    head, _, tail = name.partition("@")  # handle version identifier
    version, _, tail = tail.partition("/")  # get section after version
    return PurePosixPath(tail or head).suffix or ".js"


def resolve_module_exports_from_file(
    file: Path,
    max_depth: int,
    is_re_export: bool = False,
) -> Set[str]:
    if max_depth == 0:
        logger.warning(f"Did not resolve all exports for {file} - max depth reached")
        return set()
    elif not file.exists():
        logger.warning(f"Did not resolve exports for unknown file {file}")
        return set()

    export_names, references = resolve_module_exports_from_source(
        file.read_text(), exclude_default=is_re_export
    )

    for ref in references:
        if urlparse(ref).scheme:  # is an absolute URL
            export_names.update(
                resolve_module_exports_from_url(ref, max_depth - 1, is_re_export=True)
            )
        else:
            path = file.parent.joinpath(*ref.split("/"))
            export_names.update(
                resolve_module_exports_from_file(path, max_depth - 1, is_re_export=True)
            )

    return export_names


def resolve_module_exports_from_url(
    url: str,
    max_depth: int,
    is_re_export: bool = False,
) -> Set[str]:
    if max_depth == 0:
        logger.warning(f"Did not resolve all exports for {url} - max depth reached")
        return set()

    try:
        text = requests.get(url).text
    except requests.exceptions.ConnectionError as error:
        reason = "" if error is None else " - {error.errno}"
        logger.warning("Did not resolve exports for url " + url + reason)
        return set()

    export_names, references = resolve_module_exports_from_source(
        text, exclude_default=is_re_export
    )

    for ref in references:
        url = _resolve_relative_url(url, ref)
        export_names.update(
            resolve_module_exports_from_url(url, max_depth - 1, is_re_export=True)
        )

    return export_names


def resolve_module_exports_from_source(
    content: str, exclude_default: bool
) -> Tuple[Set[str], Set[str]]:
    names: Set[str] = set()
    references: Set[str] = set()

    if _JS_DEFAULT_EXPORT_PATTERN.search(content):
        names.add("default")

    # Exporting functions and classes
    names.update(_JS_FUNC_OR_CLS_EXPORT_PATTERN.findall(content))

    for export in _JS_GENERAL_EXPORT_PATTERN.findall(content):
        export = export.rstrip(";").strip()
        # Exporting individual features
        if export.startswith("let "):
            names.update(let.split("=", 1)[0] for let in export[4:].split(","))
        # Renaming exports and export list
        elif export.startswith("{") and export.endswith("}"):
            names.update(
                item.split(" as ", 1)[-1] for item in export.strip("{}").split(",")
            )
        # Exporting destructured assignments with renaming
        elif export.startswith("const "):
            names.update(
                item.split(":", 1)[0]
                for item in export[6:].split("=", 1)[0].strip("{}").split(",")
            )
        # Default exports
        elif export.startswith("default "):
            names.add("default")
        # Aggregating modules
        elif export.startswith("* as "):
            names.add(export[5:].split(" from ", 1)[0])
        elif export.startswith("* "):
            references.add(export[2:].split("from ", 1)[-1].strip("'\""))
        elif export.startswith("{") and " from " in export:
            names.update(
                item.split(" as ", 1)[-1]
                for item in export.split(" from ")[0].strip("{}").split(",")
            )
        elif not (export.startswith("function ") or export.startswith("class ")):
            logger.warning(f"Unknown export type {export!r}")

    names = {n.strip() for n in names}
    references = {r.strip() for r in references}

    if exclude_default and "default" in names:
        names.remove("default")

    return names, references


def _resolve_relative_url(base_url: str, rel_url: str) -> str:
    if not rel_url.startswith("."):
        return rel_url

    base_url = base_url.rsplit("/", 1)[0]

    if rel_url.startswith("./"):
        return base_url + rel_url[1:]

    while rel_url.startswith("../"):
        base_url = base_url.rsplit("/", 1)[0]
        rel_url = rel_url[3:]

    return f"{base_url}/{rel_url}"


_JS_DEFAULT_EXPORT_PATTERN = re.compile(
    r";?\s*export\s+default\s",
)
_JS_FUNC_OR_CLS_EXPORT_PATTERN = re.compile(
    r";?\s*export\s+(?:function|class)\s+([a-zA-Z_$][0-9a-zA-Z_$]*)"
)
_JS_GENERAL_EXPORT_PATTERN = re.compile(
    r"(?:^|;|})\s*export(?=\s+|{)(.*?)(?=;|$)", re.MULTILINE
)



"""
IDOM provides a series of configuration options that can be set using environment
variables or, for those which allow it, a programatic interface.
"""

from pathlib import Path
from tempfile import TemporaryDirectory

from ._option import DeprecatedOption as _DeprecatedOption
from ._option import Option as _Option


IDOM_DEBUG_MODE = _Option(
    "IDOM_DEBUG_MODE",
    default=False,
    mutable=False,
    validator=lambda x: bool(int(x)),
)
"""This immutable option turns on/off debug mode

The string values ``1`` and ``0`` are mapped to ``True`` and ``False`` respectively.

When debug is on, extra validation measures are applied that negatively impact
performance but can be used to catch bugs during development. Additionally, the default
log level for IDOM is set to ``DEBUG``.
"""

IDOM_CHECK_VDOM_SPEC = _Option(
    "IDOM_CHECK_VDOM_SPEC",
    default=IDOM_DEBUG_MODE.current,
    mutable=False,
    validator=lambda x: bool(int(x)),
)
"""This immutable option turns on/off checks which ensure VDOM is rendered to spec

The string values ``1`` and ``0`` are mapped to ``True`` and ``False`` respectively.

By default this check is off. When ``IDOM_DEBUG_MODE=1`` this will be turned on but can
be manually disablled by setting ``IDOM_CHECK_VDOM_SPEC=0`` in addition.

For more info on the VDOM spec, see here: :ref:`VDOM JSON Schema`
"""

# Because these web modules will be linked dynamically at runtime this can be temporary
_DEFAULT_WEB_MODULES_DIR = TemporaryDirectory()

IDOM_WED_MODULES_DIR: _Option[Path] = _DeprecatedOption(
    new_name="IDOM_WEB_MODULES_DIR",
    name="IDOM_WED_MODULES_DIR",
    default=Path(_DEFAULT_WEB_MODULES_DIR.name),
    validator=Path,
)
"""This has been renamed to :data:`IDOM_WEB_MODULES_DIR`"""

IDOM_WEB_MODULES_DIR = _Option(
    "IDOM_WEB_MODULES_DIR",
    default=Path(_DEFAULT_WEB_MODULES_DIR.name),
    validator=Path,
)
"""The location IDOM will use to store its client application

This directory **MUST** be treated as a black box. Downstream applications **MUST NOT**
assume anything about the structure of this directory see :mod:`idom.web.module` for a
set of publically available APIs for working with the client.
"""

IDOM_FEATURE_INDEX_AS_DEFAULT_KEY = _Option(
    "IDOM_FEATURE_INDEX_AS_DEFAULT_KEY",
    default=True,
    mutable=False,
    validator=lambda x: bool(int(x)),
)
"""Use the index of elements/components amongst their siblings as the default key.

The flag's default value is set to true. To return to legacy behavior set
``IDOM_FEATURE_INDEX_AS_DEFAULT_KEY=0``. In a future release, this flag will be removed
entirely and the indices will always be the default key.

For more information on changes to this feature flag see:
https://github.com/idom-team/idom/issues/351
"""



from __future__ import annotations

import logging
import re
import shutil
from contextlib import contextmanager
from functools import wraps
from traceback import format_exception
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    NoReturn,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlencode, urlunparse
from uuid import uuid4
from weakref import ref

from selenium.webdriver import Chrome
from selenium.webdriver.remote.webdriver import WebDriver

from idom.config import IDOM_WEB_MODULES_DIR
from idom.core.events import EventHandler, to_event_handler_function
from idom.core.hooks import LifeCycleHook, current_hook
from idom.server.prefab import hotswap_server
from idom.server.proto import ServerFactory, ServerType
from idom.server.utils import find_available_port

from .log import ROOT_LOGGER


__all__ = [
    "find_available_port",
    "create_simple_selenium_web_driver",
    "ServerMountPoint",
]


def create_simple_selenium_web_driver(
    driver_type: Type[WebDriver] = Chrome,
    driver_options: Optional[Any] = None,
    implicit_wait_timeout: float = 10.0,
    page_load_timeout: float = 5.0,
    window_size: Tuple[int, int] = (1080, 800),
) -> WebDriver:
    driver = driver_type(options=driver_options)

    driver.set_window_size(*window_size)
    driver.set_page_load_timeout(page_load_timeout)
    driver.implicitly_wait(implicit_wait_timeout)

    return driver


_Self = TypeVar("_Self", bound="ServerMountPoint[Any, Any]")
_Mount = TypeVar("_Mount")
_Server = TypeVar("_Server", bound=ServerType[Any])
_App = TypeVar("_App")
_Config = TypeVar("_Config")


class ServerMountPoint(Generic[_Mount, _Server]):
    """A context manager for imperatively mounting views to a render server when testing"""

    mount: _Mount
    server: _Server

    _log_handler: "_LogRecordCaptor"

    def __init__(
        self,
        server_type: Optional[ServerFactory[_App, _Config]] = None,
        host: str = "127.0.0.1",
        port: Optional[int] = None,
        server_config: Optional[_Config] = None,
        run_kwargs: Optional[Dict[str, Any]] = None,
        mount_and_server_constructor: "Callable[..., Tuple[_Mount, _Server]]" = hotswap_server,  # type: ignore
        app: Optional[_App] = None,
        **other_options: Any,
    ) -> None:
        self.host = host
        self.port = port or find_available_port(host, allow_reuse_waiting_ports=False)
        self._mount_and_server_constructor: "Callable[[], Tuple[_Mount, _Server]]" = (
            lambda: mount_and_server_constructor(
                server_type,
                self.host,
                self.port,
                server_config,
                run_kwargs,
                app,
                **other_options,
            )
        )

    @property
    def log_records(self) -> List[logging.LogRecord]:
        """A list of captured log records"""
        return self._log_handler.records

    def url(self, path: str = "", query: Optional[Any] = None) -> str:
        """Return a URL string pointing to the host and point of the server

        Args:
            path: the path to a resource on the server
            query: a dictionary or list of query parameters
        """
        return urlunparse(
            [
                "http",
                f"{self.host}:{self.port}",
                path,
                "",
                urlencode(query or ()),
                "",
            ]
        )

    def list_logged_exceptions(
        self,
        pattern: str = "",
        types: Union[Type[Any], Tuple[Type[Any], ...]] = Exception,
        log_level: int = logging.ERROR,
        del_log_records: bool = True,
    ) -> List[BaseException]:
        """Return a list of logged exception matching the given criteria

        Args:
            log_level: The level of log to check
            exclude_exc_types: Any exception types to ignore
            del_log_records: Whether to delete the log records for yielded exceptions
        """
        found: List[BaseException] = []
        compiled_pattern = re.compile(pattern)
        for index, record in enumerate(self.log_records):
            if record.levelno >= log_level and record.exc_info:
                error = record.exc_info[1]
                if (
                    error is not None
                    and isinstance(error, types)
                    and compiled_pattern.search(str(error))
                ):
                    if del_log_records:
                        del self.log_records[index - len(found)]
                    found.append(error)
        return found

    def __enter__(self: _Self) -> _Self:
        self._log_handler = _LogRecordCaptor()
        logging.getLogger().addHandler(self._log_handler)
        self.mount, self.server = self._mount_and_server_constructor()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.server.stop()
        logging.getLogger().removeHandler(self._log_handler)
        del self.mount, self.server
        logged_errors = self.list_logged_exceptions(del_log_records=False)
        if logged_errors:  # pragma: no cover
            raise logged_errors[0]
        return None


class LogAssertionError(AssertionError):
    """An assertion error raised in relation to log messages."""


@contextmanager
def assert_idom_logged(
    match_message: str = "",
    error_type: type[Exception] | None = None,
    match_error: str = "",
    clear_matched_records: bool = False,
) -> Iterator[None]:
    """Assert that IDOM produced a log matching the described message or error.

    Args:
        match_message: Must match a logged message.
        error_type: Checks the type of logged exceptions.
        match_error: Must match an error message.
        clear_matched_records: Whether to remove logged records that match.
    """
    message_pattern = re.compile(match_message)
    error_pattern = re.compile(match_error)

    try:
        with capture_idom_logs(yield_existing=clear_matched_records) as log_records:
            yield None
    except Exception:
        raise
    else:
        found = False
        for record in list(log_records):
            if (
                # record message matches
                message_pattern.findall(record.getMessage())
                # error type matches
                and (
                    error_type is None
                    or (
                        record.exc_info is not None
                        and record.exc_info[0] is not None
                        and issubclass(record.exc_info[0], error_type)
                    )
                )
                # error message pattern matches
                and (
                    not match_error
                    or (
                        record.exc_info is not None
                        and error_pattern.findall(
                            "".join(format_exception(*record.exc_info))
                        )
                    )
                )
            ):
                found = True
                if clear_matched_records:
                    log_records.remove(record)

        if not found:  # pragma: no cover
            _raise_log_message_error(
                "Could not find a log record matching the given",
                match_message,
                error_type,
                match_error,
            )


@contextmanager
def assert_idom_did_not_log(
    match_message: str = "",
    error_type: type[Exception] | None = None,
    match_error: str = "",
    clear_matched_records: bool = False,
) -> Iterator[None]:
    """Assert the inverse of :func:`assert_idom_logged`"""
    try:
        with assert_idom_logged(
            match_message, error_type, match_error, clear_matched_records
        ):
            yield None
    except LogAssertionError:
        pass
    else:
        _raise_log_message_error(
            "Did find a log record matching the given",
            match_message,
            error_type,
            match_error,
        )


def _raise_log_message_error(
    prefix: str,
    match_message: str = "",
    error_type: type[Exception] | None = None,
    match_error: str = "",
) -> NoReturn:
    conditions = []
    if match_message:
        conditions.append(f"log message pattern {match_message!r}")
    if error_type:
        conditions.append(f"exception type {error_type}")
    if match_error:
        conditions.append(f"error message pattern {match_error!r}")
    raise LogAssertionError(prefix + " " + " and ".join(conditions))


@contextmanager
def capture_idom_logs(
    yield_existing: bool = False,
) -> Iterator[list[logging.LogRecord]]:
    """Capture logs from IDOM

    Parameters:
        yield_existing:
            If already inside an existing capture context yield the same list of logs.
            This is useful if you need to mutate the list of logs to affect behavior in
            the outer context.
    """
    if yield_existing:
        for handler in reversed(ROOT_LOGGER.handlers):
            if isinstance(handler, _LogRecordCaptor):
                yield handler.records
                return None

    handler = _LogRecordCaptor()
    original_level = ROOT_LOGGER.level

    ROOT_LOGGER.setLevel(logging.DEBUG)
    ROOT_LOGGER.addHandler(handler)
    try:
        yield handler.records
    finally:
        ROOT_LOGGER.removeHandler(handler)
        ROOT_LOGGER.setLevel(original_level)


class _LogRecordCaptor(logging.NullHandler):
    def __init__(self) -> None:
        self.records: List[logging.LogRecord] = []
        super().__init__()

    def handle(self, record: logging.LogRecord) -> bool:
        self.records.append(record)
        return True


class HookCatcher:
    """Utility for capturing a LifeCycleHook from a component

    Example:
        .. code-block::

            hooks = HookCatcher(index_by_kwarg="key")

            @idom.component
            @hooks.capture
            def MyComponent(key):
                ...

            ... # render the component

            # grab the last render of where MyComponent(key='some_key')
            hooks.index["some_key"]
            # or grab the hook from the component's last render
            hooks.latest

        After the first render of ``MyComponent`` the ``HookCatcher`` will have
        captured the component's ``LifeCycleHook``.
    """

    latest: LifeCycleHook

    def __init__(self, index_by_kwarg: Optional[str] = None):
        self.index_by_kwarg = index_by_kwarg
        self.index: Dict[Any, LifeCycleHook] = {}

    def capture(self, render_function: Callable[..., Any]) -> Callable[..., Any]:
        """Decorator for capturing a ``LifeCycleHook`` on each render of a component"""

        # The render function holds a reference to `self` and, via the `LifeCycleHook`,
        # the component. Some tests check whether components are garbage collected, thus
        # we must use a `ref` here to ensure these checks pass once the catcher itself
        # has been collected.
        self_ref = ref(self)

        @wraps(render_function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            self = self_ref()
            assert self is not None, "Hook catcher has been garbage collected"

            hook = current_hook()
            if self.index_by_kwarg is not None:
                self.index[kwargs[self.index_by_kwarg]] = hook
            self.latest = hook
            return render_function(*args, **kwargs)

        return wrapper


class StaticEventHandler:
    """Utility for capturing the target of one event handler

    Example:
        .. code-block::

            static_handler = StaticEventHandler()

            @idom.component
            def MyComponent():
                state, set_state = idom.hooks.use_state(0)
                handler = static_handler.use(lambda event: set_state(state + 1))
                return idom.html.button({"onClick": handler}, "Click me!")

            # gives the target ID for onClick where from the last render of MyComponent
            static_handlers.target

        If you need to capture event handlers from different instances of a component
        the you should create multiple ``StaticEventHandler`` instances.

        .. code-block::

            static_handlers_by_key = {
                "first": StaticEventHandler(),
                "second": StaticEventHandler(),
            }

            @idom.component
            def Parent():
                return idom.html.div(Child(key="first"), Child(key="second"))

            @idom.component
            def Child(key):
                state, set_state = idom.hooks.use_state(0)
                handler = static_handlers_by_key[key].use(lambda event: set_state(state + 1))
                return idom.html.button({"onClick": handler}, "Click me!")

            # grab the individual targets for each instance above
            first_target = static_handlers_by_key["first"].target
            second_target = static_handlers_by_key["second"].target
    """

    def __init__(self) -> None:
        self.target = uuid4().hex

    def use(
        self,
        function: Callable[..., Any],
        stop_propagation: bool = False,
        prevent_default: bool = False,
    ) -> EventHandler:
        return EventHandler(
            to_event_handler_function(function),
            stop_propagation,
            prevent_default,
            self.target,
        )


def clear_idom_web_modules_dir() -> None:
    for path in IDOM_WEB_MODULES_DIR.current.iterdir():
        shutil.rmtree(path) if path.is_dir() else path.unlink()



from __future__ import annotations

import asyncio
import json
import logging
from asyncio import Future
from asyncio.events import AbstractEventLoop
from threading import Event
from typing import Any, Dict, Optional, Tuple, Union

from mypy_extensions import TypedDict
from sanic import Blueprint, Sanic, request, response
from sanic_cors import CORS
from websockets import WebSocketCommonProtocol

from idom.config import IDOM_WEB_MODULES_DIR
from idom.core.dispatcher import (
    RecvCoroutine,
    SendCoroutine,
    SharedViewDispatcher,
    VdomJsonPatch,
    dispatch_single_view,
    ensure_shared_view_dispatcher_future,
)
from idom.core.layout import Layout, LayoutEvent
from idom.core.proto import ComponentConstructor

from .utils import CLIENT_BUILD_DIR, threaded, wait_on_event


logger = logging.getLogger(__name__)

_SERVER_COUNT = 0


class Config(TypedDict, total=False):
    """Config for :class:`SanicRenderServer`"""

    cors: Union[bool, Dict[str, Any]]
    """Enable or configure Cross Origin Resource Sharing (CORS)

    For more information see docs for ``sanic_cors.CORS``
    """

    redirect_root_to_index: bool
    """Whether to redirect the root URL (with prefix) to ``index.html``"""

    serve_static_files: bool
    """Whether or not to serve static files (i.e. web modules)"""

    url_prefix: str
    """The URL prefix where IDOM resources will be served from"""


def PerClientStateServer(
    constructor: ComponentConstructor,
    config: Optional[Config] = None,
    app: Optional[Sanic] = None,
) -> SanicServer:
    """Return a :class:`SanicServer` where each client has its own state.

    Implements the :class:`~idom.server.proto.ServerFactory` protocol

    Parameters:
        constructor: A component constructor
        config: Options for configuring server behavior
        app: An application instance (otherwise a default instance is created)
    """
    config, app = _setup_config_and_app(config, app)
    blueprint = Blueprint(f"idom_dispatcher_{id(app)}", url_prefix=config["url_prefix"])
    _setup_common_routes(blueprint, config)
    _setup_single_view_dispatcher_route(blueprint, constructor)
    app.blueprint(blueprint)
    return SanicServer(app)


def SharedClientStateServer(
    constructor: ComponentConstructor,
    config: Optional[Config] = None,
    app: Optional[Sanic] = None,
) -> SanicServer:
    """Return a :class:`SanicServer` where each client shares state.

    Implements the :class:`~idom.server.proto.ServerFactory` protocol

    Parameters:
        constructor: A component constructor
        config: Options for configuring server behavior
        app: An application instance (otherwise a default instance is created)
    """
    config, app = _setup_config_and_app(config, app)
    blueprint = Blueprint(f"idom_dispatcher_{id(app)}", url_prefix=config["url_prefix"])
    _setup_common_routes(blueprint, config)
    _setup_shared_view_dispatcher_route(app, blueprint, constructor)
    app.blueprint(blueprint)
    return SanicServer(app)


class SanicServer:
    """A thin wrapper for running a Sanic application

    See :class:`idom.server.proto.Server` for more info
    """

    _loop: AbstractEventLoop

    def __init__(self, app: Sanic) -> None:
        self.app = app
        self._did_start = Event()
        self._did_stop = Event()
        app.register_listener(self._server_did_start, "after_server_start")
        app.register_listener(self._server_did_stop, "after_server_stop")

    def run(self, host: str, port: int, *args: Any, **kwargs: Any) -> None:
        self.app.run(host, port, *args, **kwargs)  # pragma: no cover

    @threaded
    def run_in_thread(self, host: str, port: int, *args: Any, **kwargs: Any) -> None:
        loop = asyncio.get_event_loop()

        # what follows was copied from:
        # https://github.com/sanic-org/sanic/blob/7028eae083b0da72d09111b9892ddcc00bce7df4/examples/run_async_advanced.py

        serv_coro = self.app.create_server(
            host, port, *args, **kwargs, return_asyncio_server=True
        )
        serv_task = asyncio.ensure_future(serv_coro, loop=loop)
        server = loop.run_until_complete(serv_task)
        server.after_start()
        try:
            loop.run_forever()
        except KeyboardInterrupt:  # pragma: no cover
            loop.stop()
        finally:
            server.before_stop()

            # Wait for server to close
            close_task = server.close()
            loop.run_until_complete(close_task)

            # Complete all tasks on the loop
            for connection in server.connections:
                connection.close_if_idle()
            server.after_stop()

    def wait_until_started(self, timeout: Optional[float] = 3.0) -> None:
        wait_on_event(f"start {self.app}", self._did_start, timeout)

    def stop(self, timeout: Optional[float] = 3.0) -> None:
        self._loop.call_soon_threadsafe(self.app.stop)
        wait_on_event(f"stop {self.app}", self._did_stop, timeout)

    async def _server_did_start(self, app: Sanic, loop: AbstractEventLoop) -> None:
        self._loop = loop
        self._did_start.set()

    async def _server_did_stop(self, app: Sanic, loop: AbstractEventLoop) -> None:
        self._did_stop.set()


def _setup_config_and_app(
    config: Optional[Config],
    app: Optional[Sanic],
) -> Tuple[Config, Sanic]:
    if app is None:
        global _SERVER_COUNT
        _SERVER_COUNT += 1
        app = Sanic(f"{__name__}[{_SERVER_COUNT}]")
    return (
        {
            "cors": False,
            "url_prefix": "",
            "serve_static_files": True,
            "redirect_root_to_index": True,
            **(config or {}),  # type: ignore
        },
        app,
    )


def _setup_common_routes(blueprint: Blueprint, config: Config) -> None:
    cors_config = config["cors"]
    if cors_config:  # pragma: no cover
        cors_params = cors_config if isinstance(cors_config, dict) else {}
        CORS(blueprint, **cors_params)

    if config["serve_static_files"]:
        blueprint.static("/client", str(CLIENT_BUILD_DIR))
        blueprint.static("/modules", str(IDOM_WEB_MODULES_DIR.current))

        if config["redirect_root_to_index"]:

            @blueprint.route("/")  # type: ignore
            def redirect_to_index(
                request: request.Request,
            ) -> response.HTTPResponse:
                return response.redirect(
                    f"{blueprint.url_prefix}/client/index.html?{request.query_string}"
                )


def _setup_single_view_dispatcher_route(
    blueprint: Blueprint, constructor: ComponentConstructor
) -> None:
    @blueprint.websocket("/stream")  # type: ignore
    async def model_stream(
        request: request.Request, socket: WebSocketCommonProtocol
    ) -> None:
        send, recv = _make_send_recv_callbacks(socket)
        component_params = {k: request.args.get(k) for k in request.args}
        await dispatch_single_view(Layout(constructor(**component_params)), send, recv)


def _setup_shared_view_dispatcher_route(
    app: Sanic, blueprint: Blueprint, constructor: ComponentConstructor
) -> None:
    dispatcher_future: Future[None]
    dispatch_coroutine: SharedViewDispatcher

    async def activate_dispatcher(app: Sanic, loop: AbstractEventLoop) -> None:
        nonlocal dispatcher_future
        nonlocal dispatch_coroutine
        dispatcher_future, dispatch_coroutine = ensure_shared_view_dispatcher_future(
            Layout(constructor())
        )

    async def deactivate_dispatcher(app: Sanic, loop: AbstractEventLoop) -> None:
        logger.debug("Stopping dispatcher - server is shutting down")
        dispatcher_future.cancel()
        await asyncio.wait([dispatcher_future])

    app.register_listener(activate_dispatcher, "before_server_start")
    app.register_listener(deactivate_dispatcher, "before_server_stop")

    @blueprint.websocket("/stream")  # type: ignore
    async def model_stream(
        request: request.Request, socket: WebSocketCommonProtocol
    ) -> None:
        if request.args:
            raise ValueError(
                "SharedClientState server does not support per-client view parameters"
            )
        send, recv = _make_send_recv_callbacks(socket)
        await dispatch_coroutine(send, recv)


def _make_send_recv_callbacks(
    socket: WebSocketCommonProtocol,
) -> Tuple[SendCoroutine, RecvCoroutine]:
    async def sock_send(value: VdomJsonPatch) -> None:
        await socket.send(json.dumps(value))

    async def sock_recv() -> LayoutEvent:
        return LayoutEvent(**json.loads(await socket.recv()))

    return sock_send, sock_recv
